/*
 * arch/arm/mach-spear6xx/include/mach/io.h
 *
 * IO definitions for SPEAr6xx machine family
 *
 * Copyright (C) 2009 ST Microelectronics
 * Rajeev Kumar Kumar<rajeev-dlh.kumar@st.com>
 *
 * This file is licensed under the terms of the GNU General Public
 * License version 2. This program is licensed "as is" without any
 * warranty of any kind, whether express or implied.
 */

#ifndef __MACH_IO_H
#define __MACH_IO_H

#include <plat/io.h>

#endif	/* __MACH_IO_H */

